import copy
from time import time
from typing import Tuple

import ray
import torch
import torch.nn.functional as F
import wandb
from torch import nn
from torchtyping import TensorType
from trlx.data.configs import TRLConfig
from trlx.model import register_model
from trlx.model.accelerate_ppo_model import AcceleratePPOModel
from trlx.model.nn.ppo_models import CausalLMHydraWithValueHead


class SoftEmbedding(nn.Module):
    def __init__(
        self,
        wte: nn.Embedding,
        n_tokens: int = 10,
        random_range: float = 0.5,
        initialize_from_vocab: bool = True,
    ):
        """
        appends learned embedding as prefix

        From: https://github.com/kipgparker/soft-prompt-tuning

        Args:
            wte (nn.Embedding): original transformer word embedding
            n_tokens (int, optional): number of tokens for task. Defaults to 10.
            random_range (float, optional): range to init embedding (if not initialize from vocab). Defaults to 0.5.
            initialize_from_vocab (bool, optional): initalizes from default vocab. Defaults to True.
        """
        super().__init__()
        self.wte = wte
        self.n_tokens = n_tokens
        self.padding_token_id = 50256  # used when input tensors are prefix padded
        self.learned_embedding = (
            nn.parameter.Parameter(  # dim: (n_tokens, embedding_dim)
                self.initialize_embedding(
                    wte, n_tokens, random_range, initialize_from_vocab
                )
            )
        )
        self.init_embedding = copy.deepcopy(self.learned_embedding)

    def initialize_embedding(
        self,
        wte: nn.Embedding,
        n_tokens: int = 10,
        random_range: float = 0.5,
        initialize_from_vocab: bool = True,
    ):
        """
        initializes learned embedding

        Args:
            same as __init__
        Returns:
            torch.float: initialized using original schemes
        """
        if initialize_from_vocab:
            return self.wte.weight[:n_tokens].clone().detach()
        return torch.FloatTensor(n_tokens, wte.weight.size(1)).uniform_(
            -random_range, random_range
        )

    def forward(self, tokens):
        """
        run forward pass

        Args:
            tokens (torch.long): input tokens before encoding
        Returns:
            seq_embedding (torch.float): encoding of text concatenated with learned task specifc embedding
        """
        prompt_tokens = tokens[
            :, self.n_tokens :
        ]  # dim: (batch_size, seq_length) - without soft prompt padding indices
        if self.padding_token_id in prompt_tokens[:, 0]:  # padding is applied as prefix
            seq_embedding = self.wte(
                tokens
            )  # dim: (batch_size, seq_length, embedding_dim)
            padding_tensor = torch.tensor([self.padding_token_id]).to(
                seq_embedding.device
            )

            # index in each sequence in tokens just after last prefix padding
            # this would be where the (first) soft prompt embedding should be set
            first_prompt_indices = (
                (prompt_tokens == padding_tensor).int().argmin(axis=1)
            )

            # for asserting that the first main sequence token embedding isn't modified by accident
            first_prompt_indices_full_seq = (
                (tokens == padding_tensor).int().argmin(axis=1)
            )
            first_item_idx = first_prompt_indices_full_seq[0]
            main_embedding_before_soft_prompt_assign = seq_embedding[0, first_item_idx]

            # for each batch sequence, replace embeddings at soft prompt indices with correct soft embeddings
            for batch_idx, first_prompt_idx in enumerate(first_prompt_indices.tolist()):
                # indices for assigning soft embeddings
                start = first_prompt_idx
                end = first_prompt_idx + self.n_tokens

                seq_embedding[batch_idx, start:end] = self.learned_embedding

                # debug only
                if batch_idx == 0:
                    first_main_embedding_after_soft_embedding_assign = seq_embedding[
                        0, first_item_idx
                    ]
                    assert torch.equal(
                        main_embedding_before_soft_prompt_assign,
                        first_main_embedding_after_soft_embedding_assign,
                    ), "Error: soft prompt overwrote main prompt embeddings"
        else:
            input_embedding = self.wte(prompt_tokens)
            learned_embedding = self.learned_embedding.repeat(
                input_embedding.size(0), 1, 1
            )
            seq_embedding = torch.cat([learned_embedding, input_embedding], 1)

        assert (
            seq_embedding.shape[1] == prompt_tokens.shape[1] + self.n_tokens
        ), "Number of token embeddings with soft prompts should be number of prompt tokens + number of soft tokens"

        return seq_embedding


@register_model
class AcceleratePPOSoftpromptModel(AcceleratePPOModel):
    def __init__(self, config, train_mode=True):
        # account for extra prefix tokens
        config.method.gen_kwargs["max_length"] += config.method.n_soft_tokens
        config.method.gen_kwargs["min_length"] += config.method.n_soft_tokens

        super().__init__(config)

        assert (
            config.method.n_soft_tokens > 0
        ), "Number of soft prompt tokens should be >=1"

        self.soft_dummy_token_id = 50256  # dummy token for padding soft prompts
        self.measure_soft_embedding_drift = config.method.measure_soft_embedding_drift

    def get_arch(self, config: TRLConfig):
        """
        Load model, and set Soft Prompt module for input embeddings
        """
        model = CausalLMHydraWithValueHead(
            config.model.model_path, config.model.num_layers_unfrozen
        )

        # if all layers are frozen, freeze all params. Softprompt will still be tuned
        if config.model.num_layers_unfrozen == 0:
            model.requires_grad_(False)

            if config.method.tune_v_head:
                model.v_head.requires_grad_(True)  # unfreeze value head

        # here, we setup softprompts by initializing learned softprompt embedding(s)
        # and the model's input embeddings.
        # the model will always concatenate learned softprompt embeddings as prefix to the prompt/query after it's set
        # use config option to initialize embedding from existing vocab, or random
        self.n_soft_tokens = (
            config.method.n_soft_tokens
        )  # number of prefix tokens added to prompt, with learned embeddings

        s_wte = SoftEmbedding(
            model.base_model.get_input_embeddings(),
            n_tokens=self.n_soft_tokens,
            initialize_from_vocab=config.method.initialize_from_vocab,
        )

        model.base_model.set_input_embeddings(s_wte)

        return model

    def generate(
        self,
        input_ids: TensorType["batch_size", "seq_length"],
        attention_mask: TensorType["batch_size", "seq_length"] = None,
        **kwargs,
    ):
        """
        Wraps hf's `generate` adding some specific method's defaults

        Modified to handle indices containing soft prompts
        """
        # pad for soft prompt indices (using same token as for padding)
        input_ids = torch.cat(
            [
                torch.full(
                    (input_ids.shape[0], self.n_soft_tokens), self.soft_dummy_token_id
                ).to(input_ids.device),
                input_ids,
            ],
            1,
        )
        input_ids = input_ids.to(self.accelerator.device)
        if attention_mask is not None:
            # extend for soft prompt indices (by extending mask at the end of tensor)
            attention_mask = torch.cat(
                [
                    attention_mask,
                    torch.full((attention_mask.shape[0], self.n_soft_tokens), 1).to(
                        attention_mask.device
                    ),
                ],
                1,
            )
            attention_mask = attention_mask.to(self.accelerator.device)

        kwargs = dict(self.generate_kwargs, **kwargs)

        with torch.no_grad():
            return self.accelerator.unwrap_model(self.model).generate(
                input_ids=input_ids,
                attention_mask=attention_mask,
                use_cache=False,
                **kwargs,
            )  # disable cache needed for softprompt compatibility

    def get_model_inputs(
        self,
        query_tensors: TensorType["batch_size", "query_size"],
        response_tensors: TensorType["batch_size", "response_size"],
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Used in orchestrator and loss calculation, to compute logprobs and values

        Modified to handle indices containing soft prompts
        """
        tokens = torch.cat((query_tensors, response_tensors), dim=1)
        attention_mask = (
            tokens.not_equal(self.tokenizer.pad_token_id).long().to(tokens.device)
        )
        # to handle extra softprompts, set attention at softprompt indices
        first_non_pad_indices = torch.argmax(attention_mask, dim=1)
        for batch_idx, first_non_pad_idx in enumerate(first_non_pad_indices.tolist()):
            start = first_non_pad_idx - self.n_soft_tokens
            end = first_non_pad_idx
            attention_mask[batch_idx, start:end] = 1.0

        # For a proper positional encoding in case of left padding
        position_ids = attention_mask.cumsum(-1) - 1
        position_ids.masked_fill_(attention_mask.eq(0), 0)
        return tokens, attention_mask, position_ids

    def evaluate(self):
        """
        Samples model on `eval_prompts`, logs stats with `reward_fn` or `metric_fn` if provided

        Modified to support plotting of metrics involving soft prompts
        """
        stats = {}
        all_samples = []
        generate_time = time()
        for prompts in self.eval_dataloader:
            if isinstance(prompts, torch.Tensor):
                samples = self.generate(prompts)
            else:
                samples = self.generate(**prompts)

            if isinstance(samples, tuple):
                samples, *_ = samples

            pad_token = self.tokenizer.eos_token_id if self.tokenizer else 0
            all_samples.append(
                F.pad(
                    samples,
                    (0, self.max_length - samples.shape[1]),
                    value=pad_token,
                )
            )
        stats["generate_time"] = time() - generate_time

        samples = self.accelerator.gather(torch.vstack(all_samples))

        if self.accelerator.is_main_process:
            if self.tokenizer:
                samples = self.tokenizer.batch_decode(samples, skip_special_tokens=True)

            if isinstance(samples[0], str):
                columns_data = [samples]
            else:
                columns_data = [samples.tolist()]
            columns = ["samples"]

            # in online setting, compute the reward for validation
            if self.reward_fn:
                rewards = torch.as_tensor(self.reward_fn(samples), dtype=torch.float)
                mean_reward = rewards.mean()
                columns.append("reward")
                columns_data.append(rewards)
                stats["mean_reward"] = mean_reward
                print(f"{mean_reward=}")

            # log Euclidean distance between init and current Soft Prompt embedding parameters
            if self.measure_soft_embedding_drift:
                softprompt = self.model.base_model.get_input_embeddings()
                stats["softprompt_drift_dist"] = (
                    (softprompt.init_embedding - softprompt.learned_embedding)
                    .pow(2)
                    .sum(1)
                    .sqrt()
                    .mean()
                )

            # additionally log any other metrics
            if self.metric_fn:
                metric_time = time()
                metrics = self.metric_fn(samples)
                stats["metric_time"] = time() - metric_time

                mean_metrics = {
                    f"metrics/{k}": torch.as_tensor(xs).mean(-1)
                    for k, xs in metrics.items()
                }

                stats.update(mean_metrics)

                for metric, values in metrics.items():
                    columns.append(metric)
                    columns_data.append(values)

            rows = list(zip(*columns_data))
            print(rows[0])
            if not ray.is_initialized():
                stats["samples"] = wandb.Table(columns=columns, rows=rows)

        return stats



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
import requests
import torch
from omegaconf import DictConfig, OmegaConf

from openelm.codegen.codegen_utilities import model_setup, sample, set_seed, truncate


class MutationModel(ABC):
    """Base model class for all mutation models."""

    @abstractmethod
    def generate_program(self, seed_str: str) -> dict:
        pass


@dataclass(frozen=True)
class FunctionTemplate:
    """
    A function template for a mutation model.

    Attributes:
        func_name: (str) The name of the function that we want to execute.
        import_line: (str) The import lines we add to the code.
        func_preamble: (str) The function definition, as well as potentially a
        few initial lines to generate code.
        return_line: (str) The return line we add to the end of the code.
    """

    func_name: str
    import_line: str
    func_preamble: str
    return_line: str


class PromptMutationModel(MutationModel):
    """Mutation model that uses prompts to change a seed."""

    def __init__(
        self,
        cfg,
        function_template: FunctionTemplate,
        sandbox_server="http://localhost:5000",
    ) -> None:
        if isinstance(cfg, str):
            self.cfg = OmegaConf.load(cfg)
        elif isinstance(cfg, (dict, DictConfig)):
            self.cfg = DictConfig(cfg)
        else:
            raise ValueError

        set_seed(self.cfg.seed)
        # Use RNG to rotate random seeds during inference.
        self.rng = np.random.default_rng(seed=self.cfg.seed)
        self.sandbox_server = sandbox_server
        self.model, self.tokenizer = model_setup(self.cfg)
        self.func_template: FunctionTemplate = function_template

    def generate_prompt_str(
        self,
        seed: str,
        tokenizer=None,
        batch_size=None,
        append_return=True,
        without_trunc=True,
    ) -> list[str]:
        """
        Generate a mutation string from prompting a model with a seed and function template.

        Args:
            seed: (str) the seed text.
            tokenizer: (Optional) assign only if you want to use a different
            tokenizer (default: None).
            batch_size: (Optional) override the batch size in config.
            append_return: (Optional) append a return line to the code in the end.
            without_trunc: (Optional) True if we don't apply the `truncate` function.

        Returns:
            A list of code generated by the model.
        """
        tokenizer = self.tokenizer if tokenizer is None else tokenizer
        encoding = tokenizer(
            [seed + "\n\n" + self.func_template.func_preamble],
            truncation=True,
            padding=True,
            max_length=self.cfg.gen_max_len,
            return_tensors="pt",
        )

        cfg = OmegaConf.merge(
            self.cfg,
            {"batch_size": self.cfg.batch_size if batch_size is None else batch_size},
        )
        with torch.no_grad():
            completion = sample(cfg, self.model, self.tokenizer, encoding)
        # Reset random seed
        set_seed(int(self.rng.integers(0, 1e8)))

        if without_trunc:
            truncation = completion
        else:
            truncation = [
                truncate(code, print_num=float("inf"), only_local_scope=True)
                for code in completion
            ]

        truncation = [
            self.func_template.import_line
            + "\n"
            + self.func_template.func_preamble
            + "\n"
            + code
            for code in truncation
        ]

        if append_return:
            truncation = [
                code + "\n" + self.func_template.return_line for code in truncation
            ]

        return truncation

    def generate_program(self, code: str) -> list[dict]:
        """
        Given a piece of code, do prompt mutation, call the sandbox server to execute the code and return the result.

        Args:
            code: the full code string.
        Returns:
            a numpy array (if successful) or the exception object.
        """
        results = []
        for code in self.generate_prompt_str(code):
            resp = self._get_response(code, self.cfg.timeout)
            if resp.status_code == 200:
                return_dict = json.loads(resp.text)
                self._post_process(return_dict)
                error_code = "0"
            elif resp.status_code == 500:  # Bad request
                try:
                    msg = json.loads(resp.text)
                    return_dict = {"program_str": code, "result_obj": msg["message"]}
                    error_code = msg["unsafe_execute_error_code"]
                except Exception as e:
                    return_dict = {"program_str": code, "result_obj": str(e)}
                    error_code = 6
            else:
                return_dict = {"program_str": code, "result_obj": resp.text}
                error_code = 6

            results.append({**return_dict, "error_code": error_code})

        return results

    @abstractmethod
    def _get_response(self, code: str, timeout: int) -> requests.models.Response:
        pass

    @abstractmethod
    def _post_process(self, response_dict: dict) -> dict:
        pass


class PromptMutationForSodarace(PromptMutationModel):
    def __init__(self, cfg, sandbox_server="http://localhost:5000") -> None:
        function_template = FunctionTemplate(
            func_name="make_walker",
            import_line="from .walker import walker_creator",
            func_preamble="def make_walker():\n\twc = walker_creator()\n",
            return_line="\treturn wc.get_walker()\n",
        )
        super().__init__(cfg, function_template, sandbox_server)

    def _get_response(self, code: str, timeout: int) -> requests.models.Response:
        return requests.post(
            f"{self.sandbox_server}/gen_racer",
            json={"code": code, "timeout": timeout},
            timeout=timeout,
        )

    def _post_process(self, response_dict: dict) -> dict:
        pass


class PromptMutationForImgTask(PromptMutationModel):
    def __init__(self, cfg, sandbox_server="http://localhost:5000") -> None:
        func_name = "draw"
        func_preamble = (
            f'def {func_name}():\n\t"""Draw a yellow circle.\n'
            '\t"""\n\tpic = np.zeros((32, 32, 3))\n'
        )
        function_template = FunctionTemplate(
            func_name=func_name,
            import_line="import math\nimport numpy as np",
            func_preamble=func_preamble,
            return_line="",
        )
        super().__init__(cfg, function_template, sandbox_server)

    def reset_shape(self, shape: tuple):
        self.func_preamble = f'def {self.func_name}():\n\t"""Draw a yellow circle.\n\t"""\n\tpic = np.zeros({shape})\n'

    def _get_response(self, code: str, timeout: int) -> requests.models.Response:
        return requests.post(
            f"{self.sandbox_server}/eval_imageoptim_func",
            json={"code": code, "func_name": self.func_name, "timeout": timeout},
            timeout=timeout,
        )

    def _post_process(self, response_dict: dict) -> dict:
        response_dict["result_obj"] = np.array(response_dict["result_obj"])
        return response_dict


# TODO: complete diff model (when it's available)
class DiffModel(MutationModel):
    def __init__(self, cfg) -> None:
        raise NotImplementedError()

    def generate_prompt_str(self, seed, tokenizer):
        raise NotImplementedError()

    def generate_program(self, seed_str: str) -> dict:
        raise NotImplementedError()



import itertools
import os
import re
import shutil
from typing import Iterator

import hydra
import numpy as np
import torch
from openelm.constants import SRC_PATH
from omegaconf import OmegaConf
from tqdm import tqdm

from openelm.codegen.codegen_utilities import model_setup, sample, set_seed, truncate
from openelm.codegen.codex_execute import (
    TimeoutException,
    create_tempdir,
    reliability_guard,
    swallow_io,
    time_limit,
)


def parity_reference(b1, b2, b3, b4):
    """Return binary parity of a sequence of input bits. Return 0 for even parity, 1 for odd parity."""
    bit_sum = sum([b1, b2, b3, b4])
    return bit_sum % 2


def quadratic(a, b, c, x):
    """Return quadratic: a,b,c are coefficients and x is the independent variable."""
    return a * x**2 + b * x + c


def reset_os_funcs(rmtree, rmdir, chdir):
    shutil.rmtree = rmtree
    os.rmdir = rmdir
    os.chdir = chdir


def eval_code_string(code_str: str, ground_truth: dict, timeout: int = 5):
    if len(code_str) == 0 or "def " not in code_str:
        return 6  # No code found or no function found.
    code_dct: dict = {}
    func_match = re.search(r"def (\w+)\s*\((.*?)\):", code_str)
    if func_match:
        func_name = func_match.group(1)
    else:
        return 6  # No proper function found in code.
    with create_tempdir():

        # These system calls are needed when cleaning up tempdir.
        import os
        import shutil

        rmtree = shutil.rmtree
        rmdir = os.rmdir
        chdir = os.chdir

        # Disable functionalities that can make destructive changes to the test.
        reliability_guard()
        try:
            # TODO: Check https://arxiv.org/pdf/2209.07753.pdf
            with swallow_io():
                with time_limit(timeout):
                    exec(code_str, {}, code_dct)
                    if not all(
                        [
                            code_dct[func_name](*i) == res
                            for i, res in ground_truth.items()
                        ]
                    ):
                        reset_os_funcs(rmtree, rmdir, chdir)
                        return 1  # Code runs but fails a test.
                    else:
                        reset_os_funcs(rmtree, rmdir, chdir)
                        return 0  # Passes all tests.
        except TimeoutException:
            reset_os_funcs(rmtree, rmdir, chdir)
            return 2  # Code takes too long to run.
        except RuntimeError:
            reset_os_funcs(rmtree, rmdir, chdir)
            return 3  # Code runs but crashes.
        except SyntaxError:
            reset_os_funcs(rmtree, rmdir, chdir)
            return 4  # Code does not run - syntax error.
        except TypeError:
            reset_os_funcs(rmtree, rmdir, chdir)
            return 5  # Code does not run - type error.
        except Exception:
            reset_os_funcs(rmtree, rmdir, chdir)
            return 6  # Code fails to run - other error.


def eval_completions(
    model_output: Iterator[str], task: str = "parity", timeout: int = 5
):
    """Evaluate a batch of prompt completions on a task."""
    if task == "parity":
        ground_truth = {
            i: parity_reference(*i) for i in itertools.product(range(2), repeat=4)
        }
        for completion in model_output:
            # print(completion)
            # completion = truncate(completion)
            yield eval_code_string(completion, ground_truth, timeout)
    else:
        raise ValueError(f"Unknown task: {task}")


def mutate_code(n_bugs: int = 5, task: str = "parity"):
    """Mutate code to create n bugs."""
    mutation_template = [
        "# A buggy implementation\n#!/usr/bin/python3\n",
        "",  # placeholder for the context, e.g., the buggy code
        "\n# Fixed bugs\ndef",
    ]
    if task == "parity":
        vars = ["b", "b", "b", "b", 2]
        for i in range(n_bugs):
            vars[i] = "c" if i < 4 else 3
        func_str = (
            'def parity(b1,b2,b3,b4):\n    """Return binary parity of a sequence of input bits.'
            ' Return 0 for even parity, 1 for odd parity."""\n    bit_sum = sum(['
            "{}1,{}2,{}3,{}4])\n    return bit_sum % {}".format(*vars)
        )
        mutation_template[1] = func_str
        return "".join(mutation_template)
    else:
        raise ValueError(f"Unknown task: {task}")


def run_benchmark(cfg):
    model, tokenizer = model_setup(cfg)
    mutated_str = mutate_code(n_bugs=cfg.n_bugs, task=cfg.tasks[0])
    # mutated_encoding = tokenizer([mutated_str] * cfg.gpus, truncation=True, padding=True,
    mutated_encoding = tokenizer(
        [mutated_str],
        truncation=True,
        padding=True,
        max_length=2048,
        return_tensors="pt",
    )
    input_ids_len = mutated_encoding.input_ids.shape
    text = []
    for i in range(input_ids_len[1]):
        text.append(tokenizer.batch_decode(mutated_encoding.input_ids[:, i]))
    num_batches = cfg.n_trials // cfg.batch_size
    for i in tqdm(range(num_batches), desc=f"Running benchmark with {cfg.n_bugs} bugs"):
        set_seed(torch.random.seed())
        completions = sample(cfg, model, tokenizer, mutated_encoding, add_def=True)
        truncations = map(truncate, completions)
        if i == 0:
            eval_results = np.fromiter(
                eval_completions(truncations, task=cfg.tasks[0], timeout=cfg.timeout),
                dtype=np.byte,
            )
        else:
            eval_results = np.vstack(
                (
                    eval_results,
                    np.fromiter(
                        eval_completions(
                            truncations, task=cfg.tasks[0], timeout=cfg.timeout
                        ),
                        dtype=np.byte,
                    ),
                )
            )
    corr_cnt = np.count_nonzero(eval_results == 0)
    print(f"Number of bugs: {cfg.n_bugs}")
    print(
        f"Result: {corr_cnt} successful completions in {cfg.n_trials} trials, {(corr_cnt / cfg.n_trials) * 100}%"
    )


# Load hydra config from yaml files and command line arguments.
@hydra.main(
    config_path=str(SRC_PATH / "config"),
    config_name="benchmark_cfg",
    version_base="1.2",
)
def main(cfg):
    print("----------------- Config ---------------")
    print(OmegaConf.to_yaml(cfg))
    print("-----------------  End -----------------")
    run_benchmark(cfg)


if __name__ == "__main__":
    main()



import json

from flask import Flask, abort, jsonify, request
from numpy import ndarray

from .environments.walker.walk_creator import Walker
from .utils import sandbox_unsafe_execute

app = Flask(__name__)


def bad_request(message, **kwargs):
    return {"message": message, **kwargs}, 500


def generate_racer(code_str, timeout):
    try:
        execution_result = sandbox_unsafe_execute(code_str, "make_walker", timeout)
    except:
        return bad_request(f"failed to execute code", unsafe_execute_error_code=6)  # 6: Other errors.
    if isinstance(execution_result, Walker):
        sodaracer_dict: dict = execution_result.to_dict()
        if execution_result.validate():
            return {
                "program_str": code_str,
                "result_obj": sodaracer_dict,
            }, 200
        else:
            # 1: Code runs but fails a test (valid walker).
            return bad_request(f"invalid walker", walker=execution_result.to_dict(), unsafe_execute_error_code=1)
    elif isinstance(execution_result, int):
        return bad_request(
            f"failed sandbox_unsafe_execute", unsafe_execute_error_code=execution_result
        )
    else:
        return bad_request(f"not walker", unsafe_execute_error_code=1)  # 1: Code runs but fails a test (valid walker).


@app.route("/gen_racer", methods=["POST"])
def gen_racer():
    req_json = request.get_json()
    if isinstance(req_json["code"], list):
        return [
            generate_racer(code, req_json["timeout"])[0] for code in req_json["code"]
        ], 200
    else:
        return generate_racer(req_json["code"], req_json["timeout"])


@app.route("/eval_imageoptim_func", methods=["POST"])
def evaluate_function():
    req_json = request.get_json()
    try:
        execution_result = sandbox_unsafe_execute(
            req_json["code"], req_json["func_name"], req_json["timeout"]
        )
        if isinstance(execution_result, ndarray):
            return {"program_str": req_json["code"],
                    "result_obj": execution_result.tolist().__repr__()}, 200
        elif isinstance(execution_result, int):
            return bad_request(
                f"failed sandbox_unsafe_execute", unsafe_execute_error_code=execution_result
            )
        else:
            bad_request(f"not image", unsafe_execute_error_code=1)  # 1: Code runs but fails a test (valid image).
    except:
        return bad_request(f"failed to execute code", unsafe_execute_error_code=6)  # 6: Other errors.



from enum import Enum
from typing import Any

import Box2D.Box2D as b2

# omg globals 🤮
# TODO: Figure out a better way than a global, otherwise my code standards guy will be more disappointed in me.
staticBodyCount: int = 0


class EntityType(Enum):
    Circle = 0
    Polygon = 1
    Rectangle = 2


class Entity:
    @staticmethod
    def circle_entity(
        body_id: str, x: float, y: float, radius: float, angle: float
    ) -> dict:
        """
        Create a circle entity.

        Args:
            body_id (str): Unique body id.
            x (float): Center x coordinate.
            y (float): Center y coordinate.
            radius (float): Radius of circle.
            angle (float): Angle of circle.

        Returns:
            dict: A dictionary specifying the circle entity.
        """
        entity: dict[str, Any] = {
            "type": EntityType.Circle,
            "bodyID": body_id,
            "x": x,
            "y": y,
            "radius": radius,
            "angle": angle,
        }
        return entity

    @staticmethod
    def polygon_entity(
        body_id: str,
        x_scale: float,
        y_scale: float,
        points: list[b2.b2Vec2],
        angle: float,
    ) -> dict:
        """
        Create a polygon entity.

        Args:
            body_id (str): Unique body id.
            x_scale (float): Scale of x coordinate.
            y_scale (float): Scale of y coordinate.
            points (list[b2.b2Vec2]): List of points in the polygon.
            angle (float): Angle of polygon.

        Returns:
            dict: A dictionary specifying the polygon entity.
        """
        entity: dict[str, Any] = {
            "type": EntityType.Polygon,
            "bodyID": body_id,
            "x": x_scale,
            "y": y_scale,
            "angle": angle,
            "polyPoints": [
                {"x": point.x, "y": point.y} for point in points
            ]
        }
        return entity

    @staticmethod
    def rectangle_entity(
        body_id: str,
        x_scale: float,
        y_scale: float,
        half_width: float,
        half_height: float,
        angle: float,
    ):
        """
        Create a rectangle entity.

        Args:
            body_id (str): Unique body id.
            x_scale (float): Scale of x coordinate.
            y_scale (float): Scale of y coordinate.
            half_width (float): Half width of rectangle.
            half_height (float): Half height of rectangle.
            angle (float): Angle of rectangle.
        """
        entity: dict[str, Any] = {
            "type": EntityType.Rectangle,
            "bodyID": body_id,
            "x": x_scale,
            "y": y_scale,
            "angle": angle,
            "halfWidth": half_width,
            "halfHeight": half_height,
        }
        return entity


class BodyInformation:
    # {
    # public:
    #     BodyInformation();
    #     //Keep an ID present for each body chunk
    #     std::string GenomeID;
    #
    #     //Required Nodes inside the structure
    #     std::vector<b2Vec2> nodes;
    #
    #     //All connections required for the sodaracer
    #     std::vector<b2Vec2> connections;
    #
    #     //this is for mapping between an index and a
    #     //connection number (required for pruning)
    #     std::map<int,int> indexToConnection;
    #
    #     //Is LEO used in the generation of this body
    #     //(important for output tracking)
    #     bool useLEO;
    #
    # private:
    #     ~BodyInformation();
    # };
    def __init__(self):
        self.genome_id: str = ""
        self.nodes = list()
        self.connections = list()
        self.index_to_connection = dict()
        self.useLEO = False


class PhysicsId:
    def __init__(self, pid: str):
        self.__id = pid

    def id(self):
        return self.__id


class DistanceAccessor(b2.b2DistanceJoint):
    # It was either this or access the name mangled
    # __getLength from the b2DistanceJoint so...
    getLength = b2._swig_new_instance_method(b2._Box2D.b2DistanceJoint___GetLength)
    setLength = b2._swig_new_instance_method(b2._Box2D.b2DistanceJoint___SetLength)


class Muscle:
    def __init__(
        self, muscle_id: str, joint: b2.b2Joint, phase: float, amplitude: float
    ):
        # store the id, joint, phase, and amplitude -- pretty simple!
        self.muscle_id = muscle_id
        self.joint = joint
        self.phase = phase
        self.amplitude = amplitude

        # pull the original length from our distance joint
        self.o_length = DistanceAccessor.getLength(
            joint
        )  # ((b2DistanceJoint*)joint)->GetLength();

    # getters to retrieve inner variables
    def id(self) -> str:
        return self.muscle_id

    def get_joint(self) -> b2.b2Joint:
        return self.joint

    def get_original_length(self) -> float:
        return self.o_length

    def get_phase(self) -> float:
        return self.phase

    def get_amplitude(self) -> float:
        return self.amplitude


class Bone:
    def __init__(self, muscle_id: str, joint: b2.b2Joint):
        self.bone_id = muscle_id
        self.joint = joint

    def get_joint(self) -> b2.b2Joint:
        return self.joint

    def id(self) -> str:
        return self.bone_id



from openelm.codegen.codegen_utilities import truncate


def test_truncate():
    """Test `truncate` with different non-positional arguments.

    Assumes the truncation at terminal strings works.
    """
    t1 = "def test():\n\treturn 1\ndef test2():\n\treturn 2"
    t2 = "def test():\n\treturn 1"
    t3 = "def test():\n\treturn 1\ndef test2():\n\treturn 2\nif __name__ == '__main__':\n\t test()"
    t4 = "\n\treturn 2\nif __name__ == '__main__':\n\t test()"

    assert truncate(t1) == "def test():\n\treturn 1\n"
    assert truncate(t1, def_num=2) == t1
    assert truncate(t2) == t2
    assert truncate(t3) == "def test():\n\treturn 1\n"
    assert truncate(t3, def_num=2) == t3
    assert truncate(t4, only_local_scope=True) == "\n\treturn 2\n"



from openelm.environments import (
    ImageOptim,
    Sodarace,
    image_init_args,
    sodarace_init_args,
)
from openelm.map_elites import MAPElites

ENVS_DICT = {"sodarace": Sodarace, "imageoptim": ImageOptim}
ARG_DICT = {"sodarace": sodarace_init_args, "imageoptim": image_init_args}


class ELM:
    def __init__(self, cfg, diff_model_cls=None, env_args: dict = None) -> None:
        """
        The main class of ELM.

        This class will load a diff model, an environment, and a QD algorithm
        from the passed config.

        Args:
            cfg: The config (e.g. OmegaConf who uses dot to access members).
            diff_model_cls: (Optional) The class of diff model. One can apply alternative models here for comparison.
            env_args: (Optional) The argument dict for Environment.
        """
        self.cfg = cfg

        # Get the defaults if `env_args` is not specified.
        if env_args is None:
            env_args = ARG_DICT[self.cfg.env_name]
        env_args["config"] = self.cfg  # Override default environment config

        # Override diff model if `diff_model_cls` is specified.
        if diff_model_cls is not None:
            self.diff_model = diff_model_cls(self.cfg)
            env_args = {**env_args, "diff_model": self.diff_model}
        else:
            self.diff_model = None

        self.seed = env_args["seed"]

        self.environment = ENVS_DICT[self.cfg.env_name](**env_args)
        self.qd_algorithm = MAPElites(
            self.environment,
            n_bins=self.cfg.behavior_n_bins,
            history_length=self.cfg.evo_history_length,
        )

    def run(self) -> str:
        """
        Run the ELM algorithm to evolve the population in the environment.

        Returns:
            str: A string representing the maximum fitness genotype. The
            `qd_algorithm` class attribute will be updated.
        """
        return self.qd_algorithm.search(
            initsteps=self.cfg.evo_init_steps, totalsteps=self.cfg.evo_n_steps
        )



from openelm.environments.sodaracer.helpers import (
    BodyInformation,
    Bone,
    DistanceAccessor,
    Entity,
    EntityType,
    Muscle,
    PhysicsId,
)
from openelm.environments.sodaracer.simulator import IESoRWorld, SodaraceSimulator
from openelm.environments.sodaracer.walker import Walker

__all__ = [
    "BodyInformation",
    "Bone",
    "DistanceAccessor",
    "Entity",
    "EntityType",
    "Muscle",
    "PhysicsId",
    "IESoRWorld",
    "SodaraceSimulator",
    "Walker",
]



from openelm.environments.sodaracer.walker.walk_creator import walker_creator


def make_square(wc, x0, y0, x1, y1):
    """Make a square with top left x0,y0 and top right x1,y1."""
    j0 = wc.add_joint(x0, y0)
    j1 = wc.add_joint(x0, y1)
    j2 = wc.add_joint(x1, y1)
    j3 = wc.add_joint(x1, y0)
    return j0, j1, j2, j3


def make_walker():
    wc = walker_creator()

    # the main body is a square
    sides = make_square(wc, 0, 0, 10, 10)
    center = wc.add_joint(5, 5)

    # connect the square with distance muscles
    for k in range(len(sides) - 1):
        wc.add_muscle(sides[k], sides[k + 1])
    wc.add_muscle(sides[3], sides[0])

    # one prong of the square is a distance muscle
    wc.add_muscle(sides[3], center)

    # the other prongs from the center of the square are active
    wc.add_muscle(sides[0], center, 5.0, 0.0)
    wc.add_muscle(sides[1], center, 10.0, 0.0)
    wc.add_muscle(sides[2], center, 2.0, 0.0)

    return wc.get_walker()



from time import time
from typing import Callable

import ray
import torch
from trlx.data.accelerate_base_datatypes import PromptBatch
from trlx.data.ppo_types import PPORLElement
from trlx.model import BaseRLModel
from trlx.orchestrator import register_orchestrator
from trlx.orchestrator.ppo_orchestrator import PPOOrchestrator
from trlx.pipeline import BasePipeline
from trlx.utils import Clock
from trlx.utils.modeling import logprobs_from_logits


@register_orchestrator
class PPOSoftpromptOrchestrator(PPOOrchestrator):
    def __init__(
        self,
        model: BaseRLModel,
        pipeline: BasePipeline,
        reward_fn: Callable,
        metric_fn: Callable = None,
        chunk_size: int = 512,
    ):
        super().__init__(model, pipeline, reward_fn, metric_fn, chunk_size)
        self.n_soft_tokens = model.model.base_model.get_input_embeddings().n_tokens

    def make_experience(self, num_rollouts: int = 1024, iter_count: int = 0):
        """
        Takes `num_rollouts` prompts from `pipeline`, samples model, computes KL againts a reference model appends PPOElements to model's `store`

        Modified to handle indices containing soft prompts
        """
        ppo_rl_elements = []
        stats = {}
        clock = Clock()
        while len(ppo_rl_elements) < num_rollouts:
            # Get next batch in prompt dataset and refresh if exhausted
            try:
                batch: PromptBatch = next(self.pipeline_iterator)
            except StopIteration:
                self.pipeline_iterator = iter(self.pipeline_loader)
                batch = next(self.pipeline_iterator)

            exp_generate_time = time()
            samples = self.rl_model.generate(**batch)
            stats["exp_generate_time"] = time() - exp_generate_time

            # here, we take care to handle additional softprompt indices
            query_len = batch.input_ids.shape[1] + self.n_soft_tokens
            query_tensors = samples[:, :query_len]
            response_tensors = samples[
                :, query_len:
            ]  # ignore softprompt padding index tokens
            texts = self.rl_model.tokenizer.batch_decode(
                samples, skip_special_tokens=True
            )
            exp_score_time = time()
            scores = torch.as_tensor(self.score(texts), device=samples.device)
            stats["exp_score_time"] = time() - exp_score_time

            # store statistics of the initial rollout as reference
            if self.ref_mean is None:
                self.ref_mean, self.ref_std = scores.mean(), scores.std()
            all_scores_mean, all_scores_std = self.running.update(scores)
            stats["exp_scores_mean"] = all_scores_mean
            stats["exp_scores_std"] = all_scores_std
            stats["running_mean"] = self.running.mean
            stats["running_std"] = self.running.std

            if self.rl_model.config.method.scale_reward == "running":
                scores /= self.running.std
            elif self.rl_model.config.method.scale_reward == "ref":
                scores /= self.ref_std

            clip_reward = self.rl_model.config.method.cliprange_reward
            if clip_reward:
                scores = torch.clip(scores, -clip_reward, clip_reward)

            # Precompute logprobs, values
            all_tokens, attention_mask, position_ids = self.rl_model.get_model_inputs(
                query_tensors.to(response_tensors.device), response_tensors
            )
            with torch.no_grad():
                logits, *_, v = self.rl_model.model(
                    all_tokens, attention_mask=attention_mask, position_ids=position_ids
                )
                # TODO(dahoas): When hydra model works need to also support generation on hydra head
                if hasattr(self.rl_model.model, "frozen_head"):
                    ref_logits = self.rl_model.model.forward_hydra(
                        all_tokens,
                        attention_mask=attention_mask,
                        position_ids=position_ids,
                        return_dict=False,
                    )
                else:
                    ref_logits, _, *_ = self.ref_model(
                        all_tokens.cpu(),
                        attention_mask=attention_mask.cpu(),
                        position_ids=position_ids.cpu(),
                    )

            ref_logits = ref_logits.to(self.rl_model.accelerator.device)
            logprobs = logprobs_from_logits(logits[:, :-1, :], all_tokens[:, 1:])
            ref_logprobs = logprobs_from_logits(
                ref_logits[:, :-1, :], all_tokens[:, 1:]
            )
            start = query_tensors.size()[1] - 1
            end = query_tensors.size()[1] + response_tensors.size()[1] - 1
            all_values = v[:, start:end]
            all_logprobs = logprobs[:, start:end]
            all_ref_logprobs = ref_logprobs[:, start:end]

            # Compute rewards
            kls = all_logprobs - all_ref_logprobs
            non_score_rewards = -self.rl_model.kl_ctl.value * kls
            all_rewards = non_score_rewards.clone()
            all_rewards[:, -1] += scores.to(self.rl_model.accelerator.device)

            query_tensors = query_tensors.cpu()
            response_tensors = response_tensors.cpu()
            all_logprobs = all_logprobs.cpu()
            all_values = all_values.cpu()
            all_rewards = all_rewards.cpu()

            exp_time = clock.tick()

            new_ppo_rl_elements = [
                PPORLElement(
                    query_tensor=query_tensors[i, :],
                    response_tensor=response_tensors[i, :],
                    logprobs=all_logprobs[i, :],
                    values=all_values[i, :],
                    rewards=all_rewards[i, :],
                )
                for i in range(query_tensors.size()[0])
            ]
            ppo_rl_elements += new_ppo_rl_elements

        stats["kl_ctl_value"] = self.rl_model.kl_ctl.value
        stats["exp_time"] = exp_time

        if not ray.is_initialized():
            self.rl_model.accelerator.log(stats, step=iter_count)

        # Push samples and rewards to model's rollout storage
        self.rl_model.push_to_store(ppo_rl_elements)



import math
from dataclasses import dataclass
from itertools import combinations


@dataclass(frozen=True)
class Walker:
    joints: list[tuple[float, float]]
    muscles: list[list]

    def to_dict(self):
        return self.__dict__

    def validate(self) -> bool:
        """
        Validate the Walker.

        logic for ensuring that the Sodaracer will not break the underlying
        Box2D physics engine
            a) that the strength of muscles is limited
            b) that each joint is connected only to so many muscles
            c) that there is a minimum distance between joints

        Returns:
            bool: Whether the walker is valid or not
        """
        max_muscles_per_joint: int = 10
        max_muscle_strength: int = 10
        min_joint_distance: float = 0.1
        for m in self.muscles:
            # Check a) that the strength of muscles is limited
            if m[-1]["type"] == "muscle":
                if m[-1]["amplitude"] > max_muscle_strength:
                    return False
        # Check b) that each joint is connected only to so many muscles
        for j in self.joints:
            joint_idx: int = self.joints.index(j)
            count: int = 0
            for m in self.muscles:
                count += m[:2].count(joint_idx)
            if count > max_muscles_per_joint:
                return False
        for j1, j2 in combinations(self.joints, r=2):
            # Check c) that there is a minimum distance between joints
            if (
                math.sqrt(((j1[0] - j2[0]) ** 2 + (j1[1] - j2[1]) ** 2))
                < min_joint_distance
            ):
                return False
        return True


class walker_creator:
    """Walker Creator Referenced in ELM Paper - https://arxiv.org/abs/2206.08896 (pg.16)."""

    def __init__(self):
        self.joints: list[tuple[float, float]] = []
        self.muscles: list[list] = []

    def add_joint(self, x: float, y: float) -> tuple[float, float]:
        """Add a spring/joint to the sodaracer."""
        j: tuple[float, float] = (x, y)
        self.joints.append(j)
        return j

    def add_muscle(
        self,
        j0: tuple[float, float],
        j1: tuple[float, float],
        amplitude: float = 0.0,
        phase: float = 0.0,
    ) -> list:
        """Add muscle to sodaracer."""
        # Should raise ValueError in caller if `index` fails.
        muscle_data: list = [self.joints.index(j0), self.joints.index(j1)]
        muscle_data.append(
            {
                "type": "distance",
                "amplitude": amplitude,
                "phase": phase,
            }
        )
        if amplitude != 0.0 or phase != 0.0:
            muscle_data[2]["type"] = "muscle"
        self.muscles.append(muscle_data)
        return muscle_data

    def get_walker(self) -> Walker:
        """Python dictionary with keys such as 'joints' and 'muscles'."""
        return Walker(self.joints, self.muscles)



from dataclasses import dataclass

from hydra.core.config_store import ConfigStore
from omegaconf import MISSING


@dataclass
class ConfigClass:
    model: str = MISSING
    checkpoints_dir: str = MISSING
    epochs: int = MISSING
    batch_size: int = MISSING
    fp16: bool = MISSING
    cuda: bool = MISSING
    gpus: int = MISSING
    seed: int = MISSING
    deterministic: bool = MISSING
    top_p: float = MISSING
    temp: float = MISSING
    timeout: float = MISSING
    gen_max_len: int = MISSING
    evo_init_steps: int = MISSING
    evo_n_steps: int = MISSING
    behavior_n_bins: int = MISSING
    evo_history_length: int = MISSING
    evaluation_steps: int = MISSING
    pad_token: str = MISSING
    env_name: str = MISSING
    run_name: str = MISSING


cs = ConfigStore.instance()
cs.store(name="elm_cfg", node=ConfigClass)



"""This is the benchmark test of diff models."""
import functools
import itertools
import multiprocessing as mp
import re
from typing import Union

import hydra
import torch
from benchmarks import eval_code_string, parity_reference
from omegaconf import OmegaConf
from tqdm import tqdm
from transformers import AutoConfig, AutoModelForCausalLM, AutoTokenizer

from openelm.codegen.codegen_utilities import set_seed
from openelm.constants import SRC_PATH
from openelm.utils.diff_eval import apply_diff, split_diff

parity_test_data = {
    i: parity_reference(*i) for i in itertools.product(range(2), repeat=4)
}


def eval_completions(
    codes: Union[str, list[str]], task: str = "parity", timeout: int = 5
) -> Union[int, list[int]]:
    """
    Evaluate (a batch of) the modified codes on a task.
    Args:
        codes: either a string or a batch of strings. The code(s) to be evaluated.
        task: (Optional) the task to be performed.
        timeout: (Optional) the timeout (in seconds).
    Returns:
        either the status code of the string or a list of status codes of the batch
        of strings (depending on whether `codes` is batched).
    """
    if task == "parity":
        _codes = codes if isinstance(codes, list) else [codes]
        results = []
        for code in _codes:
            results.append(eval_code_string(code, parity_test_data, timeout))
        if isinstance(codes, list):
            # Batch evaluation returns the batch
            return results
        else:
            # Single evaluation returns a single result
            return results[0]
    else:
        raise ValueError(f"Unknown task: {task}")


def mutate_code(n_bugs: int = 5, task: str = "parity") -> tuple:
    """
    Mutate code to create n bugs. Output the prompt in diff format.
    Args:
        n_bugs: number of bugs to introduce (from 1 to 5).
        task: (Optional) the task to be performed.
    Returns:
        mutated_code, function_string
    """
    mutation_template = [
        f"<NME> {task}.py\n<BEF> ",
        "",  # placeholder for the context, e.g., the buggy code
        "\n<MSG> Fixed bugs",
    ]
    if task == "parity":
        variables = ["b", "b", "b", "b", 2]
        for i in range(n_bugs):
            variables[i] = "c" if i < 4 else 3
        func_str = (
            'def parity(b1,b2,b3,b4):\n    """Return binary parity of a sequence of input bits.'
            ' Return 0 for even parity, 1 for odd parity."""\n    bit_sum = sum(['
            "{}1,{}2,{}3,{}4])\n    return bit_sum % {}".format(*variables)
        )
        mutation_template[1] = func_str
        return "".join(mutation_template), func_str
    else:
        raise ValueError(f"Unknown task: {task}")


@hydra.main(
    config_path=str(SRC_PATH / "config"),
    config_name="benchmark_diff_cfg",
    version_base="1.2",
)
def main(cfg):
    print("----------------- Config ---------------")
    print(OmegaConf.to_yaml(cfg))
    print("-----------------  End -----------------")

    device = torch.device("cuda" if cfg.cuda else "cpu")
    config = AutoConfig.from_pretrained(cfg.model)
    # Sometimes our model just fresh came out of training. Force use_cache to be true.
    config.use_cache = True
    tokenizer = AutoTokenizer.from_pretrained(cfg.model)
    if cfg.gpus > 1:
        model = torch.nn.DataParallel(
            AutoModelForCausalLM.from_pretrained(cfg.model, config=config),
            device_ids=list(range(cfg.gpus)),
        ).to(device)
        model.generate = model.module.generate
    else:
        model = AutoModelForCausalLM.from_pretrained(cfg.model, config=config).to(
            device
        )

    # Prepare the mutated codes with cfg.n_bugs bugs.
    mutated_str, function_str = mutate_code(n_bugs=cfg.n_bugs, task=cfg.tasks[0])
    mutated_encoding = tokenizer(
        [mutated_str],
        return_tensors="pt",
    ).to(device)
    token_len = mutated_encoding["input_ids"].shape[1]

    # Generate codes
    num_batches = cfg.n_trials // cfg.batch_size
    end_of_diff = re.compile("\n[^ +-@]+")
    codes = []
    for _ in tqdm(range(num_batches), desc=f"Running benchmark with {cfg.n_bugs} bugs"):
        set_seed(torch.random.seed())
        tokens = model.generate(
            **mutated_encoding,
            do_sample=True,
            num_return_sequences=cfg.batch_size,
            temperature=cfg.temp,
            max_length=token_len + cfg.gen_max_len,
            top_p=cfg.top_p,
            pad_token_id=cfg.pad_token,
        )
        texts = tokenizer.batch_decode(tokens)
        for text in texts:
            # split the diff text according to <NME>, <BEF>, <MSG>, <DFF>.
            parsed = split_diff(text)
            # truncate the diff hunk at the first line not starting with " ", "+", "-", or "@".
            if parsed and all(
                [s in parsed for s in ["name", "file", "message", "diff"]]
            ):
                diff_hunk = end_of_diff.split(parsed["diff"])[0]
                codes.append(apply_diff(function_str, diff_hunk))
            else:
                # Invalid format. No patching.
                codes.append(function_str)

    # Evaluate the codes in threads (most importantly, separate reliability_guard in separate
    # threads as it would otherwise disable things in this current thread).
    # We apply diff patch loosely:
    #   1. it ignores the line numbers;
    #   2. it ignores invalid lines (not starting with " ", "+" or "-" and not being "@@ ... @@").
    with mp.Pool(processes=cfg.num_process) as pool:
        eval_fn = functools.partial(
            eval_completions, task=cfg.tasks[0], timeout=cfg.timeout
        )
        results = list(pool.map(eval_fn, codes))

    corr_cnt = sum([r == 0 for r in results])
    print(f"Number of bugs: {cfg.n_bugs}\n")
    print(f"Mutated code to be fixed:\n{function_str}\n")
    print(
        f"Result: {corr_cnt} successful completions in {cfg.n_trials} trials, {(corr_cnt / cfg.n_trials) * 100}%"
    )


if __name__ == "__main__":
    main()



import numpy as np

from openelm.environments.sodaracer.walker.CPPN_mutable import query_cppn
from openelm.environments.sodaracer.walker.walk_creator import walker_creator


def make_walker():
    wc = walker_creator()

    def connect(x1, y1, x2, y2):
        if ((x1 - x2) ** 2 + (y1 - y2) ** 2) > 4.5:
            return False
        return True

    def amp(x1, y1, x2, y2):
        return max(abs(x1 - x2), abs(y1 - y2))

    def phase(x1, y1, x2, y2):
        return np.sign(x1)

    _ = query_cppn(wc, 8, 3, 1.5, connect, amp, phase)

    return wc.get_walker()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os
import pathlib
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
from importlib.metadata import version as importlib_version

import sphinx_rtd_theme  # noqa: F401

if "READTHEDOCS" in os.environ:
    src_folder = pathlib.Path(__file__).resolve().parent.parent.parent / "src"
    sys.path.append(str(src_folder))

    print("Detected running on ReadTheDocs")
    print(f"Added {src_folder} to sys.path")
# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "OpenELM"
copyright = "2022, CarperAI"
author = "CarperAI"
release = version = '.'.join(importlib_version('openelm').split('.')[:3])

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.


extensions = [
    "sphinx_rtd_theme",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx_autodoc_typehints",
]
# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["tests", "logs", "archive", "checkpoints"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_theme_options = {
    "logo_only": True,
    "display_version": True,
}

html_favicon = "images/openelm_favicon.ico"



"""
This module gives an example of how to run the main ELM class.

It uses the hydra library to load the config from the
config/elm_sodarace_cfg.yaml file.

This config file demonstrates an example of running ELM with the Sodarace
environment, a 2D physics-based environment in which robots specified by
Python dictionaries are evolved over.

"""
import hydra
from omegaconf import OmegaConf

from openelm import ELM
from openelm.constants import SRC_PATH


# Load hydra config from yaml files and command line arguments.
@hydra.main(
    config_path=str(SRC_PATH / "config"), config_name="elm_sodarace_cfg", version_base="1.2"
)
def main(cfg):
    print("----------------- Config ---------------")
    print(OmegaConf.to_yaml(cfg))
    print("-----------------  End -----------------")
    elm = ELM(cfg)
    print("Best Sodaracer: ", elm.run())


if __name__ == "__main__":
    main()



from dataclasses import dataclass

import trlx
from datasets import load_dataset

# to register the added softprompt model, and supported orchestrator, need to import here
from model.accelerate_ppo_softprompt_model import AcceleratePPOSoftpromptModel
from orchestrator.ppo_softprompt_orchestrator import PPOSoftpromptOrchestrator
from transformers import pipeline
from trlx.data.configs import TRLConfig
from trlx.data.method_configs import register_method
from trlx.model.nn.ppo_models import PPOConfig


@dataclass
@register_method
class PPOSoftpromptConfig(PPOConfig):
    n_soft_tokens: int = None
    initialize_from_vocab: bool = True  # of softprompt
    tune_v_head: bool = True  # set in case whole model is frozen (except softprompt)
    measure_soft_embedding_drift: bool = True  # for debugging purposes


if __name__ == "__main__":
    sentiment_fn = pipeline("sentiment-analysis", "lvwerra/distilbert-imdb", device=-1)

    def reward_fn(samples):
        outputs = sentiment_fn(samples, return_all_scores=True)
        sentiments = [output[1]["score"] for output in outputs]
        return sentiments

    # Take few words off of movies reviews as prompts
    imdb = load_dataset("imdb", split="train+test")
    prompts = [" ".join(review.split()[:4]) for review in imdb["text"]]

    config = TRLConfig.load_yaml("configs/ppo_softprompt_config.yml")

    model = trlx.train(
        "lvwerra/gpt2-imdb",
        reward_fn=reward_fn,
        prompts=prompts,
        eval_prompts=["I don't know much about Hungarian underground"] * 64,
        config=config,
    )

    print("DONE!")



import math
import string
from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar, Union

import numpy as np
from omegaconf import DictConfig, OmegaConf

from openelm.diff_model import PromptMutationForImgTask, PromptMutationForSodarace
from openelm.environments.sodaracer import SodaraceSimulator

Phenotype = Optional[np.ndarray]


def ackley(x: np.ndarray) -> np.ndarray:
    d = x.shape[-1]
    a = 5
    b = 0.1

    o1 = -a * np.exp(-b * np.sqrt(np.sum(x**2, axis=1) / d))
    o2 = -np.exp(np.sum(np.cos(math.tau * x) / d, axis=1))

    return -(a + math.exp(1) + o1 + o2)


class Genotype(ABC):
    def __str__(self) -> str:
        raise NotImplementedError


GenoType = TypeVar("GenoType", bound=Genotype)


class BaseEnvironment(ABC, Generic[GenoType]):
    def __init__(self) -> None:
        self.genotype_space: np.ndarray

    @abstractmethod
    def random(self, **kwarg) -> list[GenoType]:
        raise NotImplementedError

    @abstractmethod
    def mutate(self, x: GenoType, **kwarg) -> list[GenoType]:
        raise NotImplementedError

    @abstractmethod
    def fitness(self, x: GenoType) -> float:
        raise NotImplementedError

    @abstractmethod
    def to_behavior_space(self, x: GenoType) -> Optional[Phenotype]:
        raise NotImplementedError

    @property
    def max_fitness(self) -> int:
        return 0

    @property
    # [starts, endings) of search intervals
    def behavior_space(self) -> np.ndarray:
        return self.genotype_space

    @property
    def behavior_ndim(self) -> int:
        return self.behavior_space.shape[1]

    @staticmethod
    def _load_config(config) -> DictConfig:
        if isinstance(config, str):
            return OmegaConf.load(config)
        elif isinstance(config, (dict, DictConfig)):
            return DictConfig(config)
        else:
            raise ValueError


class ArrayGenotype(Genotype, np.ndarray):
    def __new__(cls, input_array):
        obj = np.asarray(input_array).view(cls)
        return obj

    def __str__(self) -> str:
        return f'({", ".join(map(str, np.asarray(self)))})'


# find all local maxima of a multimodal function
class FunctionOptim(BaseEnvironment[ArrayGenotype]):
    def __init__(self, ndim=2):
        self.genotype_ndim = ndim
        self.genotype_space = np.repeat([[-4, 4]], self.genotype_ndim, axis=0).T

    def random(self, **kwarg) -> list[ArrayGenotype]:
        return [ArrayGenotype(np.random.uniform(*self.genotype_space))]

    def mutate(self, x: ArrayGenotype, **kwarg) -> list[ArrayGenotype]:
        x = x.copy()
        ix = np.random.randint(self.genotype_ndim)
        x[ix] = x[ix] + np.random.uniform(-1, 1)
        return [x]

    def fitness(self, x: ArrayGenotype) -> float:
        return ackley(x[None])[0]

    def to_behavior_space(self, x: ArrayGenotype) -> Phenotype:
        return np.asarray(x)


class ImageGeneration(Genotype):
    """Genotype for generated images."""

    def __init__(self, program_str: str, result_obj: dict, error_code: bool):
        self.program_str = program_str
        self.result_obj = result_obj
        self.error_code = error_code
        self.valid = self.validate()

    def __str__(self) -> str:
        if self.valid:
            assert isinstance(self.result_obj, np.ndarray)
            return str(self.result_obj.reshape((-1, 3)).mean(axis=0).astype(int))
        else:
            return ""

    def validate(self) -> bool:
        return (
            isinstance(self.result_obj, np.ndarray)
            and len(self.result_obj.shape) == 3
            and self.result_obj.shape[2] == 3
        )

    def three_channel_average(self) -> Phenotype:
        """
        Average RGB channels.

        Assume the input is of shape (height, width, channel), and we average each channel to get (channel,)
        """
        # Code with invalid return -> return a `None` Phenotype.
        if self.valid:
            assert isinstance(self.result_obj, np.ndarray)
            return np.average(self.result_obj.reshape((-1, 3)), axis=0)
        else:
            return None


class ImageOptim(BaseEnvironment[ImageGeneration]):
    """
    Mutate programs that return images.

    Fitness is simply the absolute difference between the returning
    image and the target image. To map into the behavior space,
    if behavior_mode=="3-channel", the image will be divided into blocks (specified in `block_size`), and average
    values of RGB channels in each block will be put together as a point in the behavior space (average-pooling).
    """

    default_diff_model_cls = PromptMutationForImgTask
    # Record different definitions of behavior spaces in a dict. Feel free to add.
    behavior_mode_spec = {"3-channel": {"genotype_ndim": 3}}

    def __init__(
        self,
        seed: dict,
        config: Union[str, dict, DictConfig],
        target_img: np.ndarray,
        diff_model,
        behavior_mode: str = "3-channel",
        run_name: Optional[str] = None,
    ):
        """
        Args:
            seed: the seed dict.
            config: the config file path or dict.
            target_img: the target image.
            diff_model: the diff model (or alternatives).
            behavior_mode: (Optional) a string indicating the way an individual is mapped into behavior space.
            run_name: (Optional) override the run_name in config.
        """
        if isinstance(seed, dict):
            self.seed = ImageGeneration(**seed)
        else:
            raise TypeError

        self.config = self._load_config(config)
        if run_name is not None:
            self.config.run_name = run_name

        self.target_img = target_img
        self.shape = target_img.shape

        if diff_model is None:
            self.diff_model = self.default_diff_model_cls(self.config)
        else:
            self.diff_model = diff_model

        self.behavior_mode = behavior_mode
        self.genotype_ndim: int = self.behavior_mode_spec[self.behavior_mode][
            "genotype_ndim"
        ]
        self.genotype_space = np.repeat([[0, 255]], self.genotype_ndim, axis=0).T

    def generate_program(self, code: str) -> list[ImageGeneration]:
        """
        Call LM to generate a new program and run it.

        Returns:
            An ImageGeneration object containing the code, the resulting image
            and the error code.
        """
        return [
            ImageGeneration(**generated)
            for generated in self.diff_model.generate_program(code)
        ]

    def random(self) -> list[ImageGeneration]:
        """
        Randomly generate a batch of codes and evaluate their outputs.

        Returns:
            a tuple of the code string and the returning result (None if there is error).
        """
        new_images = self.generate_program(self.seed.program_str)
        return new_images

    def mutate(self, x: ImageGeneration) -> list[ImageGeneration]:
        """
        Randomly mutate a batch of codes from a given individual and evaluate their outputs.

        Args:
            x: the individual to be mutated.

        Returns:
            a tuple of the code string and the returning result (None if there is error).
        """
        new_images = self.generate_program(x.program_str)
        return new_images

    def fitness(self, x: ImageGeneration) -> float:
        if not x.valid or x.result_obj.shape != self.shape:
            return -np.inf
        return -np.abs(x.result_obj - self.target_img).sum()

    def to_behavior_space(self, x: ImageGeneration) -> Optional[Phenotype]:
        if self.behavior_mode == "3-channel":
            return x.three_channel_average()
        return None


class StringArrayGenotype(ArrayGenotype):
    def __str__(self) -> str:
        x: np.ndarray = np.round(self)
        return "".join(
            string.ascii_letters[ix]
            for ix in np.clip(x.astype(int), 0, len(string.ascii_letters) - 1)
        )


class MatchString(BaseEnvironment[StringArrayGenotype]):
    # find a string by mutating one character at a time

    def __init__(self, target: str):
        self.alphabet = string.ascii_letters

        self.target = np.array([self.alphabet.index(ch) for ch in target])
        self.genotype_ndim = self.target.shape[0]
        self.genotype_space = np.repeat(
            [[0, len(self.alphabet)]], self.genotype_ndim, axis=0
        ).T

    def random(self, **kwarg) -> list[StringArrayGenotype]:
        return [StringArrayGenotype(np.random.uniform(*self.genotype_space))]

    def mutate(self, x: StringArrayGenotype, **kwarg) -> list[StringArrayGenotype]:
        x = x.copy()
        ix = np.random.randint(self.genotype_ndim)
        x[ix] = x[ix] + np.random.uniform(-5, 5)
        return [x]

    def fitness(self, x: StringArrayGenotype) -> float:
        return -np.abs(x - self.target).sum()

    def to_behavior_space(self, x: StringArrayGenotype) -> Phenotype:
        return np.asarray(x)


class Sodaracer(Genotype):
    def __init__(self, program_str: str, result_obj: dict, error_code: bool):
        """
        The Sodaracer genotype.

        Args:
            program_str: the string for the original code.
            result_obj: the dict of sodaracer.
            error_code: whether the code executes in the sandbox.
        """
        self.program_str = program_str
        self.result_obj = result_obj
        self.error_code = error_code

        # Check whether the Sodaracer is valid.
        if self.error_code == 0:
            try:
                # Test the Sodaracer by actually invoking all the necessary simulations/evaluations.
                self.simulator = SodaraceSimulator(body=self.result_obj)
                self.morphology = self.simulator.morphology
                self.evaluate(0)
                self.valid = True
            except Exception as e:
                self.valid = False
        else:
            self.valid = False

    def evaluate(self, timesteps: int) -> float:
        return self.simulator.evaluate(timesteps)

    def __str__(self) -> str:
        return self.program_str


class Sodarace(BaseEnvironment[Sodaracer]):
    default_diff_model_cls = PromptMutationForSodarace

    def __init__(
        self,
        seed: dict,
        config: Union[str, dict, DictConfig],
        diff_model,
        eval_steps: int,
        max_height: int = 1000,
        max_width: int = 1000,
        max_mass: int = 2000,
        ndim: int = 3,
        run_name: Optional[str] = None,
    ) -> None:
        """
        Sodarace environment.

        Args:
            seed: the seed dict.
            config: the config file path or dict.
            diff_model: the diff model (or alternatives).
            eval_steps: the number of steps for sodaracer evaluation.
            max_height: (Optional) the maximal height.
            max_width: (Optional) the maximal width.
            max_mass: (Optional) the maximal mass.
            ndim: (Optional) the dimension of behavior space.
            run_name: (Optional) override the run_name in config.
        """
        if isinstance(seed, dict):
            self.seed = Sodaracer(**seed)
        else:
            raise TypeError
        self.config = self._load_config(config)
        if run_name is not None:
            self.config.run_name = run_name

        if diff_model is None:
            self.diff_model = self.default_diff_model_cls(self.config)
        else:
            self.diff_model = diff_model

        self.eval_steps = eval_steps
        self.genotype_ndim = ndim
        self.genotype_space = np.array(
            [[0, max_height], [0, max_width], [0, max_mass]]
        ).T

    def generate_program(self, code: str) -> list[Sodaracer]:
        # Call LM to generate a new program and run it, returning a dict containing the program string
        # and the dict from running it.
        return [
            Sodaracer(**generated)
            for generated in self.diff_model.generate_program(code)
        ]

    def fitness(self, x: Sodaracer) -> float:
        # Call Sodaracers environment to get the fitness.
        if x.valid:
            return x.evaluate(self.eval_steps)
        else:
            return -np.inf

    def random(self) -> list[Sodaracer]:
        new_sodaracers = self.generate_program(self.seed.program_str)
        return new_sodaracers

    def mutate(self, x: Sodaracer) -> list[Sodaracer]:
        new_sodaracers = self.generate_program(x.program_str)
        return new_sodaracers

    def to_behavior_space(self, x: Sodaracer) -> Optional[Phenotype]:
        # Map from floats of h,w,m to behavior space grid cells.
        if x.valid:
            return np.array(
                [x.morphology["height"], x.morphology["width"], x.morphology["mass"]]
            ).astype(int)
        else:
            return None



from CPPN_fixed import make_walker as make_walker_cppn_fixed
from CPPN_mutable import make_walker as make_walker_cppn_mutable
from radial import make_walker as make_walker_radial
from square import make_walker as make_walker_square

# Compare to serialized square seed from paper (pg. 17) #
square_seed = make_walker_square()
square_seed_expected = {
    "joints": [(0, 0), (0, 10), (10, 10), (10, 0), (5, 5)],
    "muscles": [
        [0, 1, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
        [1, 2, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
        [2, 3, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
        [3, 0, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
        [3, 4, {"type": "distance", "amplitude": 0.0, "phase": 0.0}],
        [0, 4, {"type": "muscle", "amplitude": 5.0, "phase": 0.0}],
        [1, 4, {"type": "muscle", "amplitude": 10.0, "phase": 0.0}],
        [2, 4, {"type": "muscle", "amplitude": 2.0, "phase": 0.0}],
    ],
}


print(
    "Square Seed Serialization Passed: ", square_seed.to_dict() == square_seed_expected
)


# Print other seeds #
radial_seed = make_walker_radial()
cppn_fixed_seed = make_walker_cppn_fixed()
cppn_mutable_seed = make_walker_cppn_mutable()
print("\n Radial Seed: \n ", radial_seed)
print("\n CPPN Fixed Seed: \n ", cppn_fixed_seed)
print("\n CPPN Mutable Seed: \n ", cppn_mutable_seed)



import math

from openelm.environments.sodaracer.walker.walk_creator import walker_creator


def make_circle(wc, cx, cy, radius, num_points):
    """Approximate a circle with center (cx,cy) square with num_points points."""
    joints = []

    tot_ang = 3.14 * 2.0

    for idx in range(num_points):
        ang = (tot_ang / num_points) * idx
        x = math.cos(ang) * radius + cx
        y = math.sin(ang) * radius + cy
        joints.append(wc.add_joint(x, y))

    return joints


def make_walker():

    wc = walker_creator()

    num_points = 8
    rad = 5.0
    cx, cy = (5, 5)
    # the main body is a square
    points = make_circle(wc, cx, cy, rad, num_points)
    center = wc.add_joint(cx, cy)

    for k in range(num_points):
        wc.add_muscle(points[k], points[(k + 1) % num_points])
        wc.add_muscle(points[k], center, float(k) / num_points, float(k) / num_points)

    return wc.get_walker()



import os
import re
import shutil

from openelm.sandbox.server.codex_execute import (
    TimeoutException,
    create_tempdir,
    reliability_guard,
    swallow_io,
    time_limit,
)


def reset_os_funcs(rmtree, rmdir, chdir):
    shutil.rmtree = rmtree
    os.rmdir = rmdir
    os.chdir = chdir


def sandbox_unsafe_execute(code_str: str, func_name: str, timeout: int = 5):
    if len(code_str) == 0 or "def " not in code_str:
        print("No code found or no function found.")
        return 6  # No code found or no function found.
    code_dct: dict = {}
    func_match = re.search(r"def (\w+)\s*\((.*?)\):", code_str)
    if not func_match:
        print("No proper function found in code.")
        return 6  # No proper function found in code.
    with create_tempdir():

        # These system calls are needed when cleaning up tempdir.
        import os
        import shutil

        rmtree = shutil.rmtree
        rmdir = os.rmdir
        chdir = os.chdir

        # Disable functionalities that can make destructive changes to the test.
        reliability_guard()

        try:
            # TODO: Check https://arxiv.org/pdf/2209.07753.pdf
            with swallow_io():
                # with time_limit(timeout):
                exec(code_str, globals(), code_dct)
                for k, v in code_dct.items():
                    globals()[k] = v
                res = code_dct[func_name]()
                reset_os_funcs(rmtree, rmdir, chdir)
                return res
        except TimeoutException:
            reset_os_funcs(rmtree, rmdir, chdir)
            return 2  # Code takes too long to run.
        except RuntimeError:
            reset_os_funcs(rmtree, rmdir, chdir)
            return 3  # Code runs but crashes.
        except SyntaxError:
            reset_os_funcs(rmtree, rmdir, chdir)
            return 4  # Code does not run - syntax error.
        except TypeError:
            reset_os_funcs(rmtree, rmdir, chdir)
            return 5  # Code does not run - type error.
        except Exception:
            reset_os_funcs(rmtree, rmdir, chdir)
            print("Code fails to run - other error.", code_dct)
            return 6  # Code fails to run - other error.



from dataclasses import dataclass

import trlx
from datasets import load_dataset

# to register the added softprompt model, and supported orchestrator, need to import here
from model.accelerate_ppo_softprompt_model import AcceleratePPOSoftpromptModel
from orchestrator.ppo_softprompt_orchestrator import PPOSoftpromptOrchestrator
from transformers import AutoTokenizer, pipeline
from trlx.data.configs import TRLConfig
from trlx.data.method_configs import register_method
from trlx.model.nn.ppo_models import PPOConfig


@dataclass
@register_method
class PPOSoftpromptConfig(PPOConfig):
    n_soft_tokens: int = None
    initialize_from_vocab: bool = True  # of softprompt
    tune_v_head: bool = True  # set in case whole model is frozen (except softprompt)


if __name__ == "__main__":
    tokenizer = AutoTokenizer.from_pretrained("lvwerra/gpt2-imdb")
    config = TRLConfig.load_yaml("configs/ppo_softprompt_config.yml")

    max_gen_length = config.method.gen_kwargs[
        "max_length"
    ]  # set reward as 0 if max length is reached.

    def reward_fn(samples):
        samples_tokenized = tokenizer(samples)
        samples_token_ids = samples_tokenized.data["input_ids"]
        reward = [(len(item_ids) / max_gen_length) for item_ids in samples_token_ids]

        return reward  # list of scalar reward scores for each response

    # Take few words off of movies reviews as prompts
    imdb = load_dataset("imdb", split="train+test")
    prompts = [" ".join(review.split()[:4]) for review in imdb["text"]]

    model = trlx.train(
        "lvwerra/gpt2-imdb",
        reward_fn=reward_fn,
        prompts=prompts,
        eval_prompts=["The following is a movie quote:"] * 64,
        config=config,
    )

    print("DONE!")
